# std

# internal
from laz.cli.parser import parser
from laz.utils.load import load
from laz.utils.logging import get_logger
from laz.model.runner import Runner
from laz.utils.utils import split_args


def root():
    parser.add_argument(
        "--reverse",
        "-r",
        action="store_true",
        default=False,
        help="Run targets in reverse order.",
    )

    cli_args, laz_args = split_args()
    if len(laz_args) == 0:
        from laz.cli.subcommands.help import help

        help()
        exit(1)

    cli_args = parser.parse_args(cli_args)
    get_logger(verbosity=cli_args.verbose)
    root_node = load()
    runner = Runner(root_node, cli_args, laz_args)
    runner.run()



# std
import os
import yaml

# internal
from laz.utils.logging import get_logger
from laz.utils import log
from laz.cli.parser import parser


subparsers = parser.add_subparsers(required=True)

parser_init = subparsers.add_parser("init")


def init():
    args = parser.parse_args()
    get_logger(verbosity=args.verbose)

    if os.path.isfile("laz.yml"):
        log.error("laz.yml exists")
        exit(1)

    data = yaml.safe_dump(
        {
            "env": {
                "example": "value",
            },
            "targets": {
                "dev": None,
            },
            "actions": {
                "example": "echo $example",
                "hello": "echo {{ target.name }}",
            },
        },
        sort_keys=False,
    ).replace("null", "")

    with open("laz.yml", "w") as fh:
        fh.write(data)



# std
from typing import Dict, List

# internal
from laz.model.path import Path
from laz.plugins.plugin import Plugin


class GroupsPlugin(Plugin):
    @property
    def groups(self) -> Dict[str, List[str]]:
        return self.context.data.get("groups") or {}

    @property
    def path(self) -> Path:
        return Path(
            self.context.data["path"],
            **self.context.data["laz"],
        )

    def before_all(self):
        path = self.path
        base = path.base
        groups = self.groups
        if base in groups:
            new_base = "{" + ",".join(self.groups[base]) + "}"
            path.base = new_base
            self.context.push({"path": path.value})



# std
import argparse
from typing import List

# internal
from laz.utils import log
from laz.utils.contexts import in_dir
from laz.model.node import Node
from laz.model.path import Path
from laz.model.resolver import Resolver
from laz.model.configuration import Configuration
from laz.model.target import Target
from laz.model.act import Act
from laz.model.action import Action
from laz.plugins.plugin import PLUGINS


class Runner:
    def __init__(self, root_node: Node, cli_args: argparse.Namespace, args: List[str]):
        self.root_node = root_node
        config = self.root_node.configuration.data.get("laz") or {}
        self.error_on_no_targets = config["error_on_no_targets"]
        self.cli_args = cli_args
        self.root_node.configuration.push(
            {
                "path": args[0],
                "args": args[1:],
            }
        )

    @property
    def resolver(self) -> Resolver:
        path = Path(
            self.root_node.configuration.data["path"],
            **self.root_node.configuration.data["laz"],
        )
        return Resolver(self.root_node, path)

    def resolve(self) -> List[Target]:
        targets = self.resolver.resolve()
        if self.cli_args.reverse:
            targets.reverse()
        return targets

    def run(self):
        self.load_plugins(self.root_node.configuration)
        self.before_all(self.root_node.configuration)
        targets = self.resolve()
        if len(targets) == 0:
            msg = "Given path resolved to zero targets"
            if self.error_on_no_targets:
                log.error(msg)
                exit(1)
            else:
                log.warning(msg)
        for target in targets:
            self._run_target(target)
        self.after_all(self.root_node.configuration)

    def _run_target(self, target: Target):
        log.debug(f"Running target {target.id}")
        with in_dir(target.data["config"]["dirpath"]):
            args = target.data["args"]
            if len(args) == 0:
                default_action = target.data["laz"]["default_action"]
                log.debug(f"Running default action: {default_action}")
                args += [default_action]
                target.data["args"] = args
            self.before_target(target)
            act = Act.new(target, args=" ".join(target.data["args"]))
            act.act()
            self.after_target(target)

    @staticmethod
    def load_plugins(configuration: Configuration):
        from importlib import import_module

        plugins = configuration.data.get("plugins", [])
        for import_path in plugins:
            import_module(import_path)

    @staticmethod
    def before_all(configuration: Configuration):
        log.debug(f"Running before_all hooks")
        for Plugin in PLUGINS:
            try:
                plugin = Plugin(configuration)
                plugin.before_all()
            except NotImplementedError:
                pass
        hook = configuration.data.get("hooks", {}).get("before_all")
        if hook is not None:
            action = Action.new(configuration, hook)
            act = Act(configuration, action=action)
            act.act()

    @staticmethod
    def before_target(target: Target):
        log.debug(f"Running before_target hooks")
        for Plugin in PLUGINS:
            try:
                plugin = Plugin(target)
                plugin.before_target()
            except NotImplementedError:
                pass
        hook = target.data.get("hooks", {}).get("before_target")
        if hook is not None:
            action = Action.new(target, hook)
            act = Act(target, action=action)
            act.act()

    @staticmethod
    def after_target(target: Target):
        log.debug(f"Running after_target hooks")
        for Plugin in PLUGINS:
            try:
                plugin = Plugin(target)
                plugin.after_target()
            except NotImplementedError:
                pass
        hook = target.data.get("hooks", {}).get("after_target")
        if hook is not None:
            action = Action.new(target, hook)
            act = Act(target, action=action)
            act.act()

    @staticmethod
    def after_all(configuration: Configuration):
        log.debug(f"Running after_all hooks")
        for Plugin in PLUGINS:
            try:
                plugin = Plugin(configuration)
                plugin.after_all()
            except NotImplementedError:
                pass
        hook = configuration.data.get("hooks", {}).get("after_all")
        if hook is not None:
            action = Action.new(configuration, hook)
            act = Act(configuration, action=action)
            act.act()



# std
from typing import Dict, List, Union

# type definitions
AtomicData = Union[type(None), bool, int, str]
ListData = List[Union[AtomicData, "ListData", "DictData"]]
DictData = Dict[str, Union[AtomicData, ListData, "DictData"]]
Data = Union[AtomicData, ListData, DictData]



# internal
from laz.model.configuration import Configuration
from laz.model.node import Node
from laz.model.target import Target


def get_target_configuration(node: Node, target_name: str) -> Target:
    target = Target(target_name)
    target_path = f"{get_path(node)}{target_name}"
    node_configuration = get_node_configuration(node)
    target.push({"target": {"name": target_name, "path": target_path}})
    target.push(node_configuration.data)
    target.push(node_configuration.get_target(target_name).data)
    return target


def get_node_configuration(node: Node) -> Configuration:
    nodes = node.root_path()
    conf = Configuration(node.configuration.id)
    for node in nodes:
        conf.push(node.configuration.data)
    return conf


def get_path(node: Node) -> str:
    nodes = node.root_path()
    names = [node.configuration.name for node in nodes]
    return "/".join(names + [""])



class Path:
    def __init__(
        self,
        value: str,
        default_base_path: str,
        default_target: str,
        path_delimiter: str,
        **_,
    ):
        self.value = value if path_delimiter in value else f"{path_delimiter}{value}"
        self.default_base_path = default_base_path
        self.default_target = default_target
        self.path_delimiter = path_delimiter

    def __repr__(self):
        return self.value

    @property
    def base(self) -> str:
        return self.value.split(self.path_delimiter)[0] or self.default_base_path

    @base.setter
    def base(self, new_base: str):
        self.value = self.value.replace(self.base, new_base)

    @property
    def target(self) -> str:
        return self.value.split(self.path_delimiter)[-1] or self.default_target

    @property
    def base_pattern(self) -> str:
        if self.base == "":
            return "**"
        else:
            return f"**/{self.base}/**"



# internal
from laz.plugins.plugin import Plugin


class TerraformPlugin(Plugin):
    @property
    def config(self) -> dict:
        return self.context.data.get("terraform") or {}

    @property
    def hooks(self) -> dict:
        return self.context.data.get("hooks") or {}

    def before_target(self):
        self._handle_log()
        self._handle_log_path()
        self._handle_workspace()
        self._handle_backend()
        self._handle_vars()
        self._handle_hooks()

    def _handle_log(self):
        tf_log = self.config.get("log")
        if tf_log:
            self.push_env(TF_LOG=tf_log)

    def _handle_log_path(self):
        tf_log_path = self.config.get("log_path")
        if tf_log_path:
            self.push_env(TF_LOG_PATH=tf_log_path)

    def _handle_workspace(self):
        tf_workspace = self.config.get("workspace")
        if tf_workspace:
            self.push_env(TF_WORKSPACE=tf_workspace)

    def _handle_backend(self):
        tf_vars = self.config.get("backend_config") or {}
        args = [f"-backend-config={key}={val}" for key, val in tf_vars.items()]
        if args:
            self.push_env(TF_CLI_ARGS_init=" ".join(args))

    def _handle_vars(self):
        tf_vars = self.config.get("vars") or {}
        env = {f"TF_VAR_{key}": val for key, val in tf_vars.items()}
        if env:
            self.push_env(**env)

    def _handle_hooks(self):
        if self.context.data["args"][0] == "terraform":
            self._handle_terraform_hook("before")
            self._handle_terraform_hook("after")
            self._handle_terraform_subcommand_hook("before")
            self._handle_terraform_subcommand_hook("after")

    def _handle_terraform_hook(self, before_or_after: str = "before"):
        hook_name = f"{before_or_after}_terraform"
        if hook_name in self.hooks:
            hook = self.hooks[hook_name]
            if not isinstance(hook, list):
                hook = [hook]
            data = {"hooks": {f"{before_or_after}_target": hook}}
            self.context.push(data)

    def _handle_terraform_subcommand_hook(self, before_or_after: str = "before"):
        subcommand = self.context.data["args"][1]
        hook_name = f"{before_or_after}_terraform_{subcommand}"
        if hook_name in self.hooks:
            hook = self.hooks[hook_name]
            if not isinstance(hook, list):
                hook = [hook]
            data = {"hooks": {f"{before_or_after}_target": hook}}
            self.context.push(data)



# external
from botocore.exceptions import ClientError


class LazError(Exception):
    pass


class LazValueError(LazError, ValueError):
    pass


class LazTypeError(LazError, TypeError):
    pass


class LazBoto3ClientError(LazError, ClientError):
    @classmethod
    def raise_if_bad_status(cls, response: dict):
        response_metadata = response["ResponseMetadata"]
        if response_metadata["HTTPStatusCode"] >= 300:
            error_response = {
                "Code": response_metadata["HTTPStatusCode"],
            }
            operation_name = "???"
            raise LazBoto3ClientError(error_response, operation_name)


class LazRuntimeError(LazError, RuntimeError):
    pass


class LazActionError(LazRuntimeError):
    pass



# std
import sys
from typing import Optional as Opt

# internal
from laz.utils import log


def main():
    log.debug(sys.argv[1:])
    subcommand = _subcommand()
    if len(sys.argv) > 1 and sys.argv[1] in ("help", "--help", "-h"):
        from laz.cli.subcommands.help import help

        help()
    elif len(sys.argv) == 2 and (
        sys.argv[1] == "version" or sys.argv[1] == "--version"
    ):
        from laz.cli.subcommands.version import version

        version()
    elif subcommand == "init":
        from laz.cli.subcommands.init import init

        init()
    elif subcommand == "show":
        from laz.cli.subcommands.show import show

        show()
    elif subcommand == "tree":
        from laz.cli.subcommands.tree import tree

        tree()
    else:
        from laz.cli.root import root

        root()


def _subcommand() -> Opt[str]:
    for i, s in enumerate(sys.argv):
        if i > 0 and not s.startswith("-"):
            return sys.argv[i]


if __name__ == "__main__":
    sys.argv = [
        "laz",
        "-vvv",
        "pqr/dev",
        "echo",
        "{{ target.name }}",
    ]
    main()



# std
from typing import List

# internal
from laz.model.path import Path
from laz.model.matcher import Matcher
from laz.model.target import Target
from laz.model.node import Node
from laz.utils.node import get_node_configuration, get_target_configuration, get_path


class Resolver:
    def __init__(self, node: Node, path: Path):
        self.node = node
        self.path = path

    def resolve(self) -> List[Target]:
        nodes = self.resolve_nodes()
        return self.resolve_targets(nodes)

    def resolve_nodes(self) -> List[Node]:
        matcher = Matcher(self.path.base_pattern)
        nodes: List[Node] = []
        for node in self.node.nodes():
            name_path = get_path(node)
            if matcher.match(name_path):
                nodes.append(node)
        return nodes

    def resolve_targets(self, nodes: List[Node]) -> List[Target]:
        matcher = Matcher(self.path.target)
        configurations = []
        for node in nodes:
            node_configuration = get_node_configuration(node)
            for target_name in node_configuration.data.get("targets", {}).keys():
                if matcher.match(target_name):
                    target_configuration = get_target_configuration(node, target_name)
                    configurations.append(target_configuration)
        return configurations



# std
from copy import deepcopy
import json
import os
from typing import Dict, Union

# external
import boto3

# internal
from laz.utils.errors import LazValueError, LazError, LazBoto3ClientError
from laz.utils import log
from laz.utils.types import Data
from laz.utils.contexts import with_environ
from laz.utils.funcs import compact_dict
from laz.model.action import Action
from laz.model.configuration import Configuration
from laz.model.target import Target
from laz.plugins.plugin import Plugin


class AwsPlugin(Plugin):
    _secret_string_cache: Dict[str, str] = {}

    @property
    def config(self):
        return self.context.data.get("aws") or {}

    def before_all(self):
        env = {}
        aws_profile = self.config.get("profile")
        if aws_profile is not None:
            env["AWS_PROFILE"] = aws_profile
        aws_region = self.config.get("region")
        if aws_region is not None:
            env["AWS_DEFAULT_REGION"] = aws_region
        if env:
            self.push_env(**env)

    def before_target(self):
        self._handle_secrets()

    def _handle_secrets(self):
        env = {}
        for key, val in self.env.items():
            if (
                isinstance(val, dict)
                and len(val.keys()) == 1
                and "aws" in val
                and isinstance(val["aws"], dict)
            ):
                secret_config = val["aws"]
                env[key] = self._get_secret_value(secret_config)
        if env:
            data = deepcopy(self.context.data)
            for key, val in env.items():
                data["env"][key] = val
            self.context.replace(data)

    def _get_secret_value(self, secret_config: dict) -> str:
        profile_name = secret_config.get("profile_name") or self.env.get("AWS_PROFILE")
        region_name = secret_config.get("region_name") or self.env.get(
            "AWS_DEFAULT_REGION"
        )
        secret_id = secret_config.get("secret_id")
        version_id = secret_config.get("version_id")
        version_stage = secret_config.get("version_stage")
        json_key = secret_config.get("json_key")
        cache_id = "|".join(
            map(str, (profile_name, region_name, secret_id, version_id, version_stage))
        )
        if cache_id in self._secret_string_cache:
            log.debug(f"Getting cached AWS secret: {cache_id}")
            secret_string = self._secret_string_cache[cache_id]
        else:
            log.debug(f"Getting AWS secret: {secret_config}")
            session = boto3.Session(profile_name=profile_name, region_name=region_name)
            client = session.client("secretsmanager")
            response = client.get_secret_value(
                **compact_dict(
                    dict(
                        SecretId=secret_id,
                        VersionId=version_id,
                        VersionStage=version_stage,
                    )
                )
            )
            LazBoto3ClientError.raise_if_bad_status(response)
            secret_string = response["SecretString"]
            self._secret_string_cache[cache_id] = secret_string
        if json_key:
            return json.loads(secret_string)[json_key]
        else:
            return secret_string


class AwsAction(Action):
    def run(self):
        if (
            isinstance(self.run_data["aws"], dict)
            and "assume_role" in self.run_data["aws"]
        ):
            return self._assume_role()
        elif isinstance(self.run_data["aws"], dict):
            return self._arbitrary_aws_action()
        else:
            raise LazValueError(f"Invalid aws plugin action")

    def _assume_role(self):
        import boto3

        aws: dict = self.run_data["aws"]
        kwargs: Dict[str, str] = aws["assume_role"]
        if "RoleSessionName" not in kwargs:
            kwargs["RoleSessionName"] = os.environ["USER"]
        with with_environ(self.context.data.get("env", {})):
            sts = boto3.client("sts")
            response = sts.assume_role(**kwargs)
        status_code = response["ResponseMetadata"]["HTTPStatusCode"]
        if status_code >= 300:
            raise LazError(f"Assume Role Error: HTTPStatusCode {status_code}")
        credentials = response["Credentials"]
        self.env(
            AWS_ACCESS_KEY_ID=credentials["AccessKeyId"],
            AWS_SECRET_ACCESS_KEY=credentials["SecretAccessKey"],
            AWS_SESSION_TOKEN=credentials["SessionToken"],
        )

    def _arbitrary_aws_action(self):
        import boto3

        with with_environ(self.context.data.get("env") or {}):
            aws: Dict[str, Dict[str, Dict[str, str]]] = self.run_data["aws"]
            service: str = list(aws.keys())[0]
            subcommand: str = list(aws[service].keys())[0]
            kwargs: Dict[str, str] = aws[service][subcommand] or {}
            client = boto3.client(service)
            method = getattr(client, subcommand)
            response = method(**kwargs)
        response_metadata = response.pop("ResponseMetadata")
        status_code = response_metadata["HTTPStatusCode"]
        if status_code >= 300:
            log.error(str(aws))
            log.error(str(response))
            raise LazError(f"AWS action failed")
        print(json.dumps(response, indent=2))
        return response

    @classmethod
    def is_handler(cls, context: Union[Configuration, Target], run_data: Data) -> bool:
        return isinstance(run_data, dict) and "aws" in run_data



# std
from fnmatch import fnmatch
from typing import List

# external
from braceexpand import braceexpand


class Matcher:
    def __init__(self, pattern: str):
        self._patterns: List[str] = list(braceexpand(pattern))

    def match(self, string: str) -> bool:
        return any(fnmatch(string, p) for p in self._patterns)



# std
from copy import deepcopy
from typing import Union

# external
from prodict import Prodict

# internal
from laz.utils.types import Data, DictData
from laz.utils.walk import Walk


def prodictify(data: Data) -> Union[Data, Prodict]:
    data = deepcopy(data)
    walk = Walk(data, data, dict_handler=_handle_dict)
    return walk.walk()


def _handle_dict(data: DictData, _) -> Prodict:
    return Prodict.from_dict(data)



# std
from argparse import Namespace
from typing import List

# internal
from laz.model.node import Node
from laz.utils.errors import LazRuntimeError
from laz.model.runner import Runner
from laz.model.target import Target
from laz.utils.load import load
from laz.utils.logging import get_logger
from laz.cli.parser import parser

# constants
from laz.utils.node import get_path, get_node_configuration

DESCRIPTION = """Show the effective configuration of a given node or target.
"""

subparsers = parser.add_subparsers(required=True)

parser_show = subparsers.add_parser("show", description=DESCRIPTION)
parser_show.add_argument("path", type=str)


def show():

    args = parser.parse_args()
    get_logger(verbosity=args.verbose)

    root_node = load()
    runner = Runner(root_node, Namespace(reverse=False), [args.path])
    nodes = runner.resolver.resolve_nodes()
    targets = runner.resolve()

    if len(targets) >= 1:
        _show_target(targets)
    elif len(nodes) >= 1:
        _show_node(nodes)
    else:
        msg = f"""Error during show command.
Given path did not resolve to any targets or nodes.
"""
        raise LazRuntimeError(msg)


def _show_target(targets: List[Target]):

    if len(targets) > 1:
        target_paths = [t.data.target.path for t in targets]
        msg = f"""Error during show command.
Given path should not resolve to more than 1 target. Got {len(targets)}.
Resolved targets: {target_paths}
"""
        raise LazRuntimeError(msg)

    target = targets[0]
    print(target.json())


def _show_node(nodes: List[Node]):

    if len(nodes) > 1:
        node_paths = [get_path(n) for n in nodes]
        msg = f"""Error during show command.
Given path should not resolve to more than 1 node. Got {len(nodes)}.
Resolved nodes: {node_paths}
"""
        raise LazRuntimeError(msg)

    node = nodes[0]
    print(get_node_configuration(node).json())



# std
from __future__ import annotations
import os
from typing import List

# external
import yaml
from yaml.loader import SafeLoader

# internal
from laz.model.base import BaseObject
from laz.model.target import Target
from laz.utils.prodict import prodictify
from laz.utils.types import Data


class Configuration(BaseObject):
    def __init__(self, id: str, root: bool = False, **data: Data):
        super().__init__(id)
        if root:
            self.push(DEFAULT_CONFIGURATION_DATA)
        self.push(
            {
                "config": {
                    "name": self.name,
                    "filepath": self.filepath,
                    "dirpath": self.dirpath,
                    "root": root,
                }
            }
        )
        self.push(data)

    @property
    def filepath(self):
        return self.id

    @property
    def dirpath(self):
        return os.path.dirname(self.filepath)

    @property
    def name(self):
        return os.path.basename(self.dirpath)

    @property
    def target_names(self) -> List[str]:
        return list(self.data.get("targets", {}).keys())

    def get_target(self, name: str) -> Target:
        return Target(name, **self.data.get("targets", {}).get(name, {}) or {})

    @classmethod
    def deserialize(cls, id: str, serialized: str, **kwargs: Data) -> Configuration:
        data = prodictify(yaml.load(serialized, Loader=SafeLoader) or {})
        return cls(id, kwargs.pop("root", False), **data)

    @classmethod
    def load(cls, filepath: str, root: bool = False) -> Configuration:
        with open(filepath, "r") as fh:
            return Configuration.deserialize(filepath, fh.read(), root=root)


DEFAULT_CONFIGURATION_DATA = {
    "plugins": [
        "laz.plugins.alias",
        "laz.plugins.jinja",
        "laz.plugins.groups",
    ],
    "laz": {
        "default_base_path": "",
        "default_target": "default",
        "default_action": "default",
        "error_on_no_targets": False,
        "path_delimiter": "/",
        # 'continue_on_error': False,  # TODO
    },
    "env": {},
    "targets": {},
    "actions": {},
}



# std
from __future__ import annotations
from abc import abstractmethod
import os
import subprocess
from typing import List, Optional as Opt, Union

# internal
from laz.utils.errors import LazValueError
from laz.utils import log
from laz.utils.types import Data, DictData, ListData
from laz.model.configuration import Configuration
from laz.model.target import Target


class Action:
    action_types = []

    def __init__(self, context: Union[Configuration, Target], run_data: Data):
        self.context: Union[Configuration, Target] = context
        self.run_data: Data = run_data

    def __init_subclass__(cls):
        if cls.__name__ != "ShellAction":
            cls.action_types.append(cls)

    @abstractmethod
    def run(self):
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def is_handler(cls, context: Union[Configuration, Target], run_data: Data) -> bool:
        raise NotImplementedError

    @classmethod
    def new(cls, context: Union[Configuration, Target], run_data: Data) -> Action:
        if cls.__name__ != "Action":
            return cls(context, run_data)
        for action_type in cls.action_types:
            if action_type.is_handler(context, run_data):
                return action_type.new(context, run_data)
        if ShellAction.is_handler(context, run_data):
            return ShellAction.new(context, run_data)
        cls.raise_value_error(context, run_data)

    @classmethod
    def raise_value_error(
        cls, context: Union[Configuration, Target], run_data: Data, msg: Opt[str] = None
    ):
        log.error(f"target -> {context.serialize()}")
        log.error(f"run data -> {run_data}")
        msg = msg or f"Given values could not be handled as a {cls.__name__}"
        raise LazValueError(msg)


class DefaultAction(Action):
    def run(self):
        raise NotImplementedError

    @classmethod
    def is_handler(cls, context: Union[Configuration, Target], run_data: Data) -> bool:
        if isinstance(run_data, str) and run_data == "default":
            return True
        else:
            return False

    @classmethod
    def new(cls, context: Union[Configuration, Target], run_data: Data) -> Action:
        configured_actions: DictData = context.data.get("actions", {})
        if isinstance(run_data, str) and run_data in configured_actions:
            return Action.new(context, configured_actions[run_data])
        else:
            cls.raise_value_error(context, run_data, "No default action found")


class ConfiguredAction(Action):
    def run(self):
        raise NotImplementedError

    @classmethod
    def is_handler(cls, context: Union[Configuration, Target], run_data: Data) -> bool:
        configured_actions: DictData = context.data.get("actions", {})
        if isinstance(run_data, str) and run_data in configured_actions:
            return True
        else:
            return False

    @classmethod
    def new(cls, context: Union[Configuration, Target], run_data: Data) -> Action:
        configured_actions: DictData = context.data.get("actions", {})
        if isinstance(run_data, str) and run_data in configured_actions:
            return Action.new(context, configured_actions[run_data])
        else:
            cls.raise_value_error(context, run_data)


class MultipleActions(Action):
    def __init__(self, context: Union[Configuration, Target], run_data: ListData):
        super().__init__(context, run_data)
        self.actions = [Action.new(context, d) for d in self.run_data]

    def run(self):
        results = []
        for action in self.actions:
            results.append(action.run())
        return results

    @classmethod
    def is_handler(cls, context: Union[Configuration, Target], run_data: Data) -> bool:
        return isinstance(run_data, list)

    @classmethod
    def new(cls, context: Union[Configuration, Target], run_data: ListData) -> Action:
        return MultipleActions(context, run_data)


class ConditionalAction(Action):
    def run(self):
        result = Action.new(self.context, self.run_data["condition"]).run()
        if bool(result):
            if "if" in self.run_data:
                return Action.new(self.context, self.run_data["if"]).run()
        else:
            if "else" in self.run_data:
                return Action.new(self.context, self.run_data["else"]).run()

    @classmethod
    def is_handler(cls, context: Union[Configuration, Target], run_data: Data) -> bool:
        return isinstance(run_data, dict) and "condition" in run_data

    @classmethod
    def new(cls, context: Union[Configuration, Target], run_data: DictData) -> Action:
        return cls(context, run_data)


class SwitchAction(Action):
    def run(self):
        result = Action.new(self.context, self.run_data["switch"]).run()
        if result in self.run_data:
            return Action.new(self.context, self.run_data[result]).run()
        elif "default" in self.run_data:
            return Action.new(self.context, self.run_data["default"]).run()
        else:
            return

    @classmethod
    def is_handler(cls, context: Union[Configuration, Target], run_data: Data) -> bool:
        return isinstance(run_data, dict) and "switch" in run_data

    @classmethod
    def new(cls, context: Union[Configuration, Target], run_data: DictData) -> Action:
        return cls(context, run_data)


class ShellAction(Action):
    run_data: DictData

    def run(self):
        env = {
            **os.environ,
            **(self.context.data.get("env") or {}),
        }
        log.debug(self.args())
        if self.run_data.get("capture"):
            type_str = self.run_data.get("type", "str")
            result = subprocess.run(
                self.args(), env=env, capture_output=True
            ).stdout.decode()
            if self.run_data.get("strip", True):
                result = result.strip()
            result = eval(f"{type_str}({result})")
            return result
        else:
            return subprocess.run(self.args(), env=env).returncode == 0

    def args(self) -> List[str]:
        return [self._shell(), "-c", self.run_data["command"]]

    def _shell(self) -> str:
        return os.environ.get("SHELL", "bash")

    @classmethod
    def is_handler(cls, context: Union[Configuration, Target], run_data: Data) -> bool:
        if isinstance(run_data, str):
            return True
        elif isinstance(run_data, dict) and "command" in run_data:
            return True
        else:
            cls.raise_value_error(context, run_data)

    @classmethod
    def new(
        cls, context: Union[Configuration, Target], run_data: Union[str, DictData]
    ) -> Action:
        if isinstance(run_data, str):
            run_data = {"command": run_data}
        return cls(context, run_data)


class PythonAction(Action):
    def run(self):
        return eval(self.run_data["python"])

    @classmethod
    def is_handler(cls, context: Union[Configuration, Target], run_data: Data) -> bool:
        return isinstance(run_data, dict) and "python" in run_data

    @classmethod
    def new(cls, context: Union[Configuration, Target], run_data: DictData) -> Action:
        return cls(context, run_data)



# std
from typing import List

# internal
from laz.utils.logging import get_logger
from laz.cli.parser import parser
from laz.utils.load import load
from laz.model.node import Node
from laz.utils.iterator import Iterator
from laz.utils.node import get_node_configuration, get_path


subparsers = parser.add_subparsers(required=True)

parser_tree = subparsers.add_parser("tree")
parser_tree.add_argument(
    "--show-targets", "-t", action="store_true", default=False, help="Show targets."
)
parser_tree.add_argument(
    "--show-paths", "-p", action="store_true", default=False, help="Show paths."
)

# config
SHOW_TARGETS = False
SHOW_PATHS = False


def tree():
    global SHOW_TARGETS, SHOW_PATHS
    args = parser.parse_args()
    get_logger(verbosity=args.verbose)
    SHOW_TARGETS = args.show_targets
    SHOW_PATHS = args.show_paths
    root = load()
    print(f"{root.configuration.name}/")
    if args.show_targets:
        _targets(root, "", len(root.children) == 0)
    for child in Iterator(root.children):
        _tree(child.value, "", child.is_last)


def _tree(
    node: Node,
    prefix: str = "",
    last: bool = False,
):
    symbol = "└─" if last else "├─"
    path = f"({get_path(node)})" if SHOW_PATHS else ""
    print(f"{prefix}{symbol} {node.configuration.name}/ {path}")

    if SHOW_TARGETS:
        _targets(node, prefix, last)

    for child in Iterator(node.children):
        _tree(child.value, prefix + "│  ", child.is_last)


def _targets(node: Node, prefix: str, last: bool):
    prefix = prefix + "   " if last else prefix + "│  "
    for target_name in Iterator(_get_target_names(node)):
        symbol = "└─" if target_name.is_last and len(node.children) == 0 else "├─"
        print(f"{prefix}{symbol} {target_name.value}")


def _get_target_names(node: Node) -> List[str]:
    configuration = get_node_configuration(node)
    return configuration.target_names



# std
import sys
from typing import List, Optional as Opt, Tuple


def split_args(args: Opt[List[str]] = None) -> Tuple[List[str], List[str]]:
    """Split args into cli args and laz args to be forwarded"""
    args = args or sys.argv
    for i, s in enumerate(args):
        if i > 0 and not s.startswith("-"):
            return args[1:i], args[i:]
    return args[1:], []
